import laspy
import numpy as np
import pandas as pd
import sys
import math
from scipy import spatial

# read parameters from excel file 
def readfile(filepath):
    RTKDATA = pd.read_excel(filepath)
    RTK_x = RTKDATA["x"]
    RTK_y = RTKDATA["y"]
    plotname = RTKDATA["plot"]
    lasfiles = RTKDATA["lasfile"]
    branchheight = RTKDATA["branch"]
    return plotname, RTK_x, RTK_y, lasfiles, branchheight

# polar transform
 # radius
def get_radius(x, y, z, x_min, y_min, z_min):
    r = np.sqrt((x - x_min) ** 2 + (y - y_min) ** 2 + (z - z_min) ** 2)
    # r = np.sqrt((x **2+y **2+z **2))
    return r

# phi
def get_phi(x, y, x_min, y_min):
    if x == x_min:
        if y == y_min:
            a = 0
        else:
            a = math.pi / 2.0 * (abs(y) / y)
    elif x > x_min:
        a = math.atan((y - y_min) / (x - x_min))
    else:
        a = math.atan((y - y_min) / (x - x_min)) + math.pi
    return a

# zenith
def get_zenith(x, y, z, x_min, y_min, z_min):
    r = np.sqrt((x - x_min) ** 2 + (y - y_min) ** 2 + (z - z_min) ** 2)
    if r == 0:
        b = 0
    else:
        b = math.asin(abs(z - z_min) / r)
        b = math.pi / 2 - b
    return b

# transform 
def polartrans(X, Y, Z, x_min, y_min, z_min):
    radius = []
    phi = []
    zenith = []
    for x, y, z in zip(X, Y, Z):
        radius.append(get_radius(x, y, z, x_min, y_min, z_min))
        phi.append(get_phi(x, y, x_min, y_min))
        zenith.append(get_zenith(x, y, z, x_min, y_min, z_min))
    return zenith, phi, radius


# project to different girds
# Radian to degree
def tramsform(zenithR, phiR, radius):
    # Determine (0,0) points to start
    zenith_new = [0]
    phi_new = [0]
    for x in zenithR:
        zenith_new.append(math.degrees(x))
    zenith_new = np.array(zenith_new)
    for x in phiR:
        phi_new.append(math.degrees(x))
    # to numpy
    radius = np.array(radius)
    phi_new = np.array(phi_new)
    zenith_new = np.array(zenith_new)
    # round 2 
    radius = radius.round(decimals=2)
    phi_new = phi_new.round(decimals=2)
    zenith_new = zenith_new.round(decimals=2)
    return phi_new,zenith_new

def calculate(phi_new, zenith_new, constant,res):
# set resolutions
    # CONVERT TO PIXEL POSITION VALUES - Based on resolution
    x_img = (phi_new / res).astype(np.int32)  # y axis is -x in LIDAR
    y_img = (zenith_new / res).astype(np.int32)  # x axis is -y in LIDAR

    # CLIP HEIGHT VALUES
    pixel_values = 255

    # INITIALIZE EMPTY ARRAY - of the dimensions we want
    y_max = 1 + int((max(zenith_new) - min(zenith_new)) / res)
    x_max = 1 + int((max(phi_new) - min(phi_new)) / res)
    im = np.zeros([y_max, x_max], dtype=np.uint8)

    # FILL PIXEL VALUES IN IMAGE ARRAY
    im[y_img, x_img] = pixel_values
    LEAF = len(im[im == 255])
    PHI_LEN = 360 / res
    ZENITH_LEN = constant / res
    CANOPYCLOSURE = LEAF / (PHI_LEN * ZENITH_LEN)
    return CANOPYCLOSURE

def CC_extract(plotname, RTK_x, RTK_y, lasfiles, branchheight):
    for plot, orgin_x, orgin_y, lasfile, branch in zip(plotname, RTK_x, RTK_y, lasfiles, branchheight):
        print(plot)
        inFile = laspy.file.File(lasfile, mode="r")
        # read x,y,z
        x, y, z = inFile.x, inFile.y, inFile.z
        # Point cloud data index construction
        # zip x,y
        lasdata = list(zip(x, y))
        # create kd-tree
        tree = spatial.KDTree(lasdata)
        # find central point（orgin_x, orgin_y）,Point cloud within a circle with a radius of 80m
        aa = tree.query_ball_point(np.array([orgin_x, orgin_y]), 80)
        # new XYZ
        X = x[aa]
        Y = y[aa]
        Z = z[aa]
        newdata = pd.DataFrame({"x": X, "y": Y, "z": Z})
        # sort and save
        sort_data = newdata[newdata["z"] > branch]
        X = sort_data["x"]
        Y = sort_data["y"]
        Z = sort_data["z"]
        x_min = orgin_x
        y_min = orgin_y
        z_min = 1.4
        # get zenith, phi, radius
        zenith, phi, radius = polartrans(X, Y, Z, x_min, y_min, z_min)
        # zenith range
        constants = [45, 60, 75]
        for constant in constants:
            TABLE = pd.DataFrame({"zenithR": zenith, "phiR": phi})
            TABLENEW = TABLE[TABLE["zenithR"] < math.radians(constant)]
            zenithR = TABLENEW["zenithR"]
            phiR = TABLENEW["phiR"]
            # radian to degree
            phi_new, zenith_new = tramsform(zenithR, phiR, radius)
            # Calculate the canopy closure of different resolutions at the zenith angle
            res = [1, 1.5, 2]
            for res in res:
                CANOPYCLOSURE = calculate(phi_new, zenith_new, constant,res)
                print(CANOPYCLOSURE)

def main():
    # Set the txt path of the information to be printed
    outputfile = open("*/PythonTest/SYS_LAS.txt", "a")
    sys.stdout = outputfile
    # Excel File location
    filepath = "*/PythonTest/readme.xlsx"
    plotname, RTK_x, RTK_y, lasfiles, branchheight = readfile(filepath)
    CC_extract(plotname, RTK_x, RTK_y, lasfiles, branchheight)

if __name__ == '__main__':
    main()